/*

Mat 259 Project 3 3D Visualization: "Visualizing IVF Patient Journeys"
16 March 2023
Processing 4.0.1
                                                     
Author: Jenni Hutson                  
Supervisor: George Legrady                            

Purpose: To visualize trends in anonymized IVF patient steps and outcomes.
A blue line indicates that the patient had a male partner, red a female partner, and yellow no partner.

Usage: 1. A mouse left-drag will rotate the camera around the subject.
          2. A right drag will zoom in and out. 
          3. A middle-drag (command-left-drag on mac) will pan. 
          4. A double-click restores the camera to its original position. 
          5. Shift-click on a point to select that patient's through line and get a text description of their IVF treatment.
          6. Press Enter to exit selection mode.
          7. Toggle labels with the button "Labels."
          8. Select a random patient with the button "Select_Random."
*/

import peasy.*; //3D Navigation library
import controlP5.*;
import java.util.Map;
import java.util.LinkedHashMap;

PeasyCam cam;
ControlP5 cp5;

Table table;
int numRows;

int boxWidth = 800;
int boxDepth = 500;
int boxHeight = 500;

int transparency = 70;
int lessTransparency = 200;
int lowestTransparency = 20;

int[] femaleColor = {2, 63, 42};
int[] maleColor = {207, 74, 39};
int[] noPartnerColor = {62, 27, 60};
int[] gray = {0, 0, 55};

float[] zVals;
float[][] yVals;

Boolean select = false;
int rowSelected = -1;

boolean Select_Random = false;
boolean randomSelected;
boolean labels = true;

boolean showF = true;
boolean showM = true;
boolean showN = true;

int[] fArray;
int[] mArray;
int[] nArray;

int numF = 0;
int numM = 0;
int numN = 0;

final int NUM_LAYERS = 7;   //Number of Layers;

LinkedHashMap<String, Integer[]> colMap = new LinkedHashMap<String, Integer[]>();


void setup() {
  colorMode(HSB, 360, 100, 100, 255);
  //size(displayWidth, displayHeight, P3D);
  size(1500, 900, P3D);
  cp5 = new ControlP5(this);
  
  cam = new PeasyCam(this, 900);
  cam.setMaximumDistance(1100);
  
  table = loadTable("aggregatedPartners.csv", "header");
  numRows = table.getRowCount();
  println("Rows: " + numRows);
  
  Integer[] nums = {0, 40};
  colMap.put("Total eggs mixed", nums);
  Integer[] num1 = {1, 40};
  colMap.put("Total embryos created", num1);
  Integer[] num2 = {2, 40};
  colMap.put("Embryos transferred", num2);
  Integer[] num3 = {3,3};
  colMap.put("Foetal sacs with pulse", num3);
  Integer[] num4 = {4, 5};
  colMap.put("Early outcome", num4);
  Integer[] num5 = {5,40};
  colMap.put("Weeks gestation", num5);
  Integer[] num6 = {6,1};
  colMap.put("Birth outcome", num6);
  
  zVals = generateRandomVals();
  yVals = new float[NUM_LAYERS][numRows];
  setUpYVals();
  setUpPartnerArrs();
  
  println(boxWidth/(NUM_LAYERS-1));
  setGui();
}

void setUpPartnerArrs() {
  for (int row =1; row<numRows; row++) {
      int partner = checkForPartner(row);
      if (partner == 0) {numF++;}
      if (partner == 1) {numM++;}
      if (partner == 2) {numN++;}
  }
  fArray = new int[numF];
  mArray = new int[numM];
  nArray = new int[numN];
  
  int f = 0;
  int m = 0;
  int n = 0;
  for (int row =1; row<numRows; row++) {
      int partner = checkForPartner(row);
      if (partner == 0) {
        fArray[f] = row;
        f++;
      } else if (partner ==1) {
        mArray[m] = row;
        m++;
      } else {
        nArray[n] = row;
        n++;
      }
  }
  
}

void setUpYVals() {
  for (int i=0; i<NUM_LAYERS; i++) {
    float[] layerVals = new float[numRows];
    for (int row =1; row<numRows; row++) {
      float yVal = getRowVal(i, row);
      layerVals[row] = yVal;
    }
    yVals[i] = layerVals.clone();
  }
}

float getRowVal(int layer, int row) {
  switch (layer) {
    case 0:
      String rangeStr = table.getString(row, 6);
      rangeStr = rangeStr.replaceAll("\\s", "");
      float yVal = convertRangeToAvg(rangeStr);
      return yVal;
    case 1:
      String rangeStr1 = table.getString(row, 7);
      rangeStr1 = rangeStr1.replaceAll("\\s", "");
      yVal = convertRangeToAvg(rangeStr1);
      return yVal;
    case 2:
      float num = table.getFloat(row, 8);
      return num;
    case 3:
      float num1 = table.getFloat(row, 14);
      return num1;
    case 4:
      String outcome = table.getString(row, 13);
      switch (outcome){
        case "None":
        case "":
          return 0;
        case "Ectopic/Hetrotopic":
          return 1;
        case "Miscarriage":
          return 2;
        case "Biochemical Pregnancy Only":
          return 3;
        case "Intrauterine Fetal Pulsation Seen":
          return 4;
      }
    case 5:
      String heart1 = table.getString(row, 15);
      if (heart1.isEmpty()) {
        return 0;
      }
      String[] heart1Arr = heart1.split(" ");
      if (heart1Arr.length == 1) {
        return Float.valueOf(heart1);
      } else {
        return Float.valueOf(heart1Arr[2]);
      }
    case 6:
      String birth = table.getString(row, 16);
      switch (birth) {
        case "Live  Birth":
          return 1;
        default:
          return 0;
      }
    default:
      return 0;
  }
}

void draw() {
  background(0,0, 10);
  noFill();
  
  stroke(100);
  box(boxWidth, boxHeight, boxDepth);
  
  for (Map.Entry me : colMap.entrySet()) {
      Integer[] vals = colMap.get(me.getKey().toString());
      pushMatrix();
      makeBox(vals[1], vals[0]);
      popMatrix();
  }
  drawLines();
  setLabels();
}

void setLabels() {
  gui();
  if((mouseX<180) & (mouseY<180)) {
    cam.setActive(false);
  } else {
    cam.setActive(true);
  }
  if (Select_Random) {
    if (!randomSelected) {
      selectRandom();
      randomSelected = true;
    }
  } else if (randomSelected) {
    randomSelected = false;
    select = false;
    rowSelected = -1;
  }
}

void makeBox(int yScale, int i) {
  float translateX = (-boxWidth/2)+ (i*boxWidth/(NUM_LAYERS-1));
  translate(translateX, 0,0);
  
  //drawing top labels
  
  if (labels) {
    fill(gray[0],gray[1],gray[2]);
    addLabels(i);
    noFill();
  }
  
  rotateY(radians(90));
  stroke(100);
  rect(-boxDepth/2, -boxHeight/2, boxDepth, boxHeight);
  
  //(yScale-1)
  float step = boxHeight/(yScale);
  if (yScale ==1) {
    yScale = 2;
  }
  for (int j=0; j<yScale; j++) {
    pushMatrix();
    translate(-boxDepth/2, (boxHeight/2)-(step*j), 0);
    line(5,0,0,-5,0,0);
    if (labels) {
      fill(gray[0],gray[1],gray[2]);
      switch(i){
        case 0:
        case 1:
        case 2:
        case 3:
        case 5:
         text(j,-15,0,0);
         break;
        case 4:
          switch (j){
            case 0:
              text("No pregnancy",-80,0,0);
              break;
            case 1:
              text("Ectopic/Hetrotopic",-110,0,0);
              break;
            case 2:
              text("Miscarriage", -80,0,0); //switch order with bio-chem?
              break;
            case 3:
              text("Biochemical Pregnancy Only", -160,0,0); 
              break;
            case 4:
              text("Fetal Pulsation Seen", -120,0,0); 
              break;
          }
          break;
       case 6:
         switch (j) {
           case 0:
             text("No live birth", -100,0,0);
             break;
           case 1:
             text("Live birth", -80, 0,0);
         }
      }
      noFill();
    }
    popMatrix();
  }
  
  if (yScale ==2) {
    yScale = 1;
  }
  
  if (i ==0) {
    scatterData(yScale);
  } else {
    drawData(yScale, i);
  }
     
}

int checkForPartner(int row) {
  String partner = table.getString(row, 21);
  switch (partner){
    case "Female":
      return 0;
    case "Male":
      return 1;
    default:
      return 2;
  }
}

void setColor(int partner, Boolean trans, Boolean isSelected) {
  if ((rowSelected > 0 && isSelected) || !select) {
    switch (partner) {
      case 0:
        if (trans) {
          stroke(femaleColor[0], femaleColor[1], femaleColor[2], transparency);
        } else {
          stroke(femaleColor[0], femaleColor[1], femaleColor[2],lessTransparency);
        }
        break;
      case 1:
        if (trans) {
          stroke(maleColor[0], maleColor[1], maleColor[2], transparency);
        } else {
          stroke(maleColor[0], maleColor[1], maleColor[2],lessTransparency);
        }
        break;
      case 2:
        if (trans) {
          stroke(noPartnerColor[0], noPartnerColor[1], noPartnerColor[2], transparency);
        } else {
          stroke(noPartnerColor[0], noPartnerColor[1], noPartnerColor[2],lessTransparency);
        }
        break;     
    }
  } else {
    if (trans) {
      stroke(gray[0], gray[1], gray[2], lowestTransparency);
    } else {
      stroke(gray[0], gray[1], gray[2], transparency);
    }
  }
}


void drawLines() {
  Object[] keys = colMap.keySet().toArray();
  for (int row =1; row<numRows; row++){
    strokeWeight(1);
    float z = zVals[row];
    for (int layer=0; layer<NUM_LAYERS-1; layer++) {
      pushMatrix();
      float translateX = (-boxWidth/2)+ (layer*boxWidth/(NUM_LAYERS-1));
      translate(translateX, 0,0);
      float y1 = yVals[layer][row]*boxHeight/colMap.get(keys[layer])[1];
      float y2 = yVals[layer+1][row]*boxHeight/colMap.get(keys[layer+1])[1];
      int partner = checkForPartner(row);
      if (!checkPartnerOn(partner)) {
        popMatrix();
        strokeWeight(1);
        continue;
      }
      boolean isSelected = false;
      if (select && rowSelected == row) {
        strokeWeight(10);
        isSelected = true;
        String info = generateInfo(row);
        writeSummary(info);
      }
      if (y2 == 0) {
        setColor(partner, true, isSelected);
      } else {
        setColor(partner, false, isSelected);
      }
      line(0, (boxHeight/2)-y1, (boxDepth/2)-z, (boxWidth/(NUM_LAYERS-1)), (boxHeight/2)-y2, (boxDepth/2)-z);
      popMatrix();
    } 
  }
  strokeWeight(1);
}

float[] generateRandomVals() {
  float[] vals = new float[numRows];
  for (int i=0; i<numRows; i++) {
    vals[i] = random(boxDepth);
  }
  return vals;
}

boolean checkPartnerOn(int partner) {
  switch (partner) {
    case 0:
      return showF;
    case 1:
      return showM;
    case 2:
      return showN;
    default:
      return true;
  }
}

void drawData(int yScale, int layer) {
    for (int row =1; row<numRows; row++) {
    stroke(100);
    float yVal = yVals[layer][row];
    float zVal = zVals[row];
    pushMatrix();
    translate((-boxDepth/2)+zVal, (boxHeight/2) -(yVal*boxHeight/yScale), 0);
    strokeWeight(3);
    int partner = checkForPartner(row);
    if (!checkPartnerOn(partner)) {
      popMatrix();
      strokeWeight(1);
      continue;
    }
    float mouseObjectDistance = sq(mouseX-screenX(0,0,0)) + sq(mouseY-screenY(0,0,0));
    boolean isSelected = false;
    if (mouseObjectDistance < 20) {
      if (rowSelected < 0) { //a row is not already selected
        strokeWeight(10);
        rowSelected = row;
        isSelected = true;
      }
    }
    setColor(partner, false, isSelected);
    point(0,0,0);
    strokeWeight(1);
    popMatrix();
  }
}

void scatterData(int yScale) {
  //give each data point a random z position within the first square
  //y position comes from number eggs mixed
  for (int row =1; row<numRows; row++) {
    //float yVal = firstRandom[row];
    float yVal = yVals[0][row];
    float zVal = zVals[row];
    pushMatrix();
    translate((-boxDepth/2)+zVal, (boxHeight/2) -(yVal*boxHeight/yScale), 0);
    strokeWeight(3);
    int partner = checkForPartner(row);
    if (!checkPartnerOn(partner)) {
      popMatrix();
      strokeWeight(1);
      continue;
    }
    float mouseObjectDistance = sq(mouseX-screenX(0,0,0)) + sq(mouseY-screenY(0,0,0));
    boolean isSelected = false;
    if (mouseObjectDistance < 35) {
      if (rowSelected < 0) { //a row is not already selected
        strokeWeight(10);
        rowSelected = row;
        isSelected = true;
      }
    }
    setColor(partner, false, isSelected);
    point(0,0,0);
    strokeWeight(1);
    popMatrix();
  }
}

float convertRangeToAvg(String range) {
  if (range.equals("0")) {
    return 0;
  } if (range.charAt(0) == '>') {
    return 40;
  } else {
    String[] rangeArr = range.split("-");
    int low = Integer.valueOf(rangeArr[0]);
    int high = Integer.valueOf(rangeArr[1]);
    float avg = random(low, high+1);
    return avg;
  }
}

int getSelectableRows() {
  int rowNum = 0;
  if (showF) {
    rowNum += numF;
  }
  if (showM) {
    rowNum += numM;
  }
  if (showN) {
    rowNum += numN;
  }
  return rowNum;
}

int selectRow(int row) {
  if (showF) {
    if (showN) {
      if (row > numF) {
        return nArray[row-numF];
      } 
    } else if (showM) {  
      if (row > numF) {
        return mArray[row-numF];
      } 
    }
    return fArray[row];
  }
  
  if (showN) {
    if (showM) {
      if (row > numN) {
        return mArray[row-numN];
      }
    }
    return nArray[row];
  }
  
  return mArray[row];
  
}

void selectRandom() {
  int rowNums = getSelectableRows();
  if (rowNums ==0) { //no rows turned on
    return;
  }
  println("numRows: ", numRows);
  println("rowNums: ", rowNums);
  int randRow = int(random(1, rowNums));
  if (!showF || !showM || !showN) {
    randRow = selectRow(randRow);
  }
  select = true;
  rowSelected = randRow;
}

void addLabels(int i) {
  //top labels
  textSize(13);
  Object[] keys = colMap.keySet().toArray();
  String labelText = keys[i].toString();
  text(labelText, 0, -boxHeight/2, boxDepth/2);
}

String generateInfo(int row) {
  String info = "Patient #" + row;
  int eggs = int(yVals[0][row]);
  if (eggs != 0) {
    info += " had " + eggs;
    if (eggs == 1) {
      info += " egg mixed, and ";
    } else {
      info += " eggs mixed, and ";
    }
    eggs = int(yVals[1][row]);
    if (eggs == 1) {
      info += eggs + " embryo created. They";
    } else {
      info += eggs + " embryos created. They";
    }
  } 
  eggs = int(yVals[2][row]);
  info+= " had ";
  if (eggs == 1) {
    info += eggs + " embryo transferred. \n Doctors";
  } else {
    info += eggs + " embryos transferred. \n Doctors";
  }
  int pulse = int(yVals[3][row]);
  info += " found " + pulse;
  if (pulse ==1) {
    info += " foetal sac with a pulse.";
  } else {
    info += " foetal sacs with a pulse.";
  }
  int outcome = int(yVals[4][row]);
  if (pulse == 0 && outcome == 0) {
    info = endInfo(info, row);
    return info;
  }
  
  switch (outcome) {
    case 0:
      info = endInfo(info, row);
      println(info);
      return info;
    case 1:
      info += " The pregnancy was ectopic or hetrotopic.";
      break;
    case 2:
      info += " The pregnancy ended in miscarriage.";
      break;
    case 3:
      info += " The patient had a biochemical pregnancy only.";
      break;
  }
  int gest = int(yVals[5][row]);
  if (gest == 0) {
    info = endInfo(info, row);
    return info;
  }
  info += " \n The patient was pregnant for " + gest;
  info += " weeks.";
  int birth = int(yVals[5][row]);
  if (birth == 0) {
    info = endInfo(info, row);
  } else {
    info += " Patient #" + row;
    info += " had a child as a result of this treatment."; 
    int previous = table.getInt(row, 1);
    if (previous != 0) {
      info += " They had had " + previous;
      if (previous == 1) {
        info += " previous round of IVF.";
      } else {
        info += " previous rounds of IVF.";
      }
    }
  }
  return info;
}

String endInfo(String info, int row) {
  String end = info + " \n They did not have a child from this round of treatment. ";
  int previous = table.getInt(row, 1);
  if (previous != 0) {
    end += "Patient #" + row;
    end += " had had " + previous;
    if (previous == 1) {
      end += " previous round of IVF.";
    } else {
      end += " previous rounds of IVF.";
    }
  }
  return end;
}

void writeSummary(String info) {
  textAlign(CENTER);
  fill(0,0,80);
  textSize(20);
  hint(DISABLE_DEPTH_TEST);
  cam.beginHUD();
  text(info, displayWidth/2, 50);
  cam.endHUD();
  hint(ENABLE_DEPTH_TEST);
  textAlign(LEFT);
}

void setGui(){
  PFont pfont= createFont("Arial", 48, true);
  ControlFont font = new ControlFont(pfont, 12);
  
  float X = 60;
  float distance = 40;
  int index = 1;
  
  cp5.addToggle("Select_Random")
  .setPosition(X, distance*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,0,100))
  .setColorActive(color(70,100,95))
  .setColorForeground(color(70,100,100))
  .setColorBackground(color(70,100,85));
  index++;
  
  cp5.addToggle("labels")
  .setPosition(X, distance*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,0,100))
  .setColorActive(color(70,100,95))
  .setColorForeground(color(70,100,100))
  .setColorBackground(color(70,100,85));
  index++;
  
  cp5.addToggle("showF")
  .setPosition(X, distance*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,0,100))
  .setCaptionLabel("Show patients with \na female partner")
  .setColorActive(color(femaleColor[0],femaleColor[1],femaleColor[2]*1.5))
  .setColorForeground(color(femaleColor[0],femaleColor[1],femaleColor[2]*1.7))
  .setColorBackground(color(femaleColor[0],femaleColor[1],femaleColor[2]-15));
  index++;
  
  cp5.addToggle("showM")
  .setPosition(X, distance*1.1*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,0,100))
  .setCaptionLabel("Show patients with \na male partner")
  .setColorActive(color(maleColor[0],maleColor[1],maleColor[2]*1.5))
  .setColorForeground(color(maleColor[0],maleColor[1],maleColor[2]*1.7))
  .setColorBackground(color(maleColor[0],maleColor[1],maleColor[2]-15));
  index++;
  
  cp5.addToggle("showN")
  .setPosition(X, distance*1.15*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,0,100))
  .setCaptionLabel("Show patients \nwithout a partner")
  .setColorActive(color(noPartnerColor[0],noPartnerColor[1],noPartnerColor[2]*1.5))
  .setColorForeground(color(noPartnerColor[0],noPartnerColor[1],noPartnerColor[2]*1.7))
  .setColorBackground(color(noPartnerColor[0],noPartnerColor[1],noPartnerColor[2]-25));
  index++;
  
  cp5.addTextlabel("label1")
  .setText("Shift + click on a point \nto select that patient.\n Press Enter to exit selection.")
  .setPosition(X-5, 20+distance*1.2*index)
  .setFont(font)
  .setColorValue(color(0,0,100));
  index++;
  
  cp5.setAutoDraw(false);
}

void gui() {
  hint(DISABLE_DEPTH_TEST);
  cam.beginHUD();
  cp5.draw();
  cam.endHUD();
  hint(ENABLE_DEPTH_TEST);
}
